"""
* Copyright (c) 2015-2017 The Hyve B.V.
* This code is licensed under the GNU General Public License,
* version 3.
* Original Author: Ruslan Forostianov

* Contributors:
    - Jochem Bijlard
    - Laura Madrid
    - Ward Weistra

* Modified by Laura Madrid on 08/03/2017
* in order to make it compatible with transmart v16.2
"""

import requests

import google.protobuf.internal.decoder as decoder
from pandas.io.json import json_normalize

from .highdim_pb2 import HighDimHeader
from .highdim_pb2 import Row
import urllib

from ..tm_api_base import TransmartAPIBase


class TransmartV1(TransmartAPIBase):
    """ Connect to tranSMART using Python. """

    def __init__(self, host, user=None, password=None, print_urls=False):
        """
        Create the python transmart client by providing user credentials.

        :param host: a transmart URL (e.g. http://transmart-test.thehyve.net)
        :param user: if not given, it asks for it.
        :param password: if not given, it asks for it.
        :param print_urls: print the url of handles being used.
        """
        super().__init__(host, user, password, print_urls)

    def get_observations(self, study=None, patientSet=None, as_dataframe=True, hal=False):
        """
        Get observations, from the main table in the transmart data model.

        :param study: studyID
        :param patientSet: patient set id
        :param as_dataframe: If True (default), convert json response to dataframe
        :param hal: ?
        :return: dataframe or direct json
        """
        url = '%s/studies/%s/observations' % (self.host, study)

        observations = self._get_json(url, hal=hal)

        if as_dataframe:
            observations = json_normalize(observations)

        return observations

    def get_patients(self, study=None, patientSet=None, as_dataframe=True, hal=False):
        """
        Get patients.

        :param study: studyID
        :param patientSet: patient set id
        :param as_dataframe: If True (default), convert json response to dataframe
        :param hal: ?
        :return: dataframe or direct json
        """
        raise ValueError("Function not implemented in Python client for API V1")

    def get_studies(self, as_dataframe=True, hal=False):
        """
        Get all studies.

        :param as_dataframe: If True (default), convert json response to dataframe
        :param hal: ?
        :return: dataframe or direct json
        """
        url = '%s/studies' % (self.host)

        studies = self._get_json(url, hal=hal)

        if as_dataframe:
            studies = json_normalize(studies['studies'])

        return studies

    def get_concepts(self, study, hal=False):
        url = '%s/studies/%s/concepts/' % (self.host, study)
        return self._get_json(url, hal=hal)

    def get_hd_node_data(self, study, node_name, projection='all_data', genes=None):
        """
        Parameters
        ----------
        node_name: string
           Name of the leaf node
        projection : string
           Possible values: default_real_projection, zscore, log_intensity, all_data (default)
        genes: list of strings
            Gene names. e.g. 'TP53', 'AURCA'
        """
        concepts = self.get_concepts(study, hal=True)
        found_condepts_hrefs = []
        for t in concepts['_embedded']['ontology_terms']:
            if t['type'] == 'HIGH_DIMENSIONAL' and t['name'] == node_name:
                found_condepts_hrefs.append(t['_links']['self']['href'])
        hd_node_url = '%s%s/highdim' % (self.host, found_condepts_hrefs[0])
        hd_node_meta = self._get_json(hd_node_url)
        hd_data_type_name = hd_node_meta['dataTypes'][0]['name']
        hd_node_data_url = '%s?projection=%s&dataType=%s' % (
            hd_node_url, projection, hd_data_type_name)
        if genes is not None:
            hd_node_data_url = hd_node_data_url + \
                '&' + urllib.parse.urlencode({'dataConstraints': {'genes': [{'names': genes}]}})
        hd_data = self._get_protobuf(hd_node_data_url)
        return hd_data

    def _get_json_post(self, url, hal=False):

        if self.print_urls:
            print(url)

        headers = {}
        headers['Accept'] = 'application/%s;charset=UTF-8' % ('hal+json' if hal else 'json')
        if self.access_token is not None:
            headers['Authorization'] = 'Bearer ' + self.access_token
        r = requests.post(url, headers=headers)
        r.raise_for_status()
        return r.json()

    def _get_json(self, url, hal=False):

        if self.print_urls:
            print(url)

        headers = {}
        headers['Accept'] = 'application/%s;charset=UTF-8' % ('hal+json' if hal else 'json')
        if self.access_token is not None:
            headers['Authorization'] = 'Bearer ' + self.access_token

        r = requests.get(url, headers=headers)
        r.raise_for_status()
        return r.json()

    def _parse_protobuf(self, data):
        hdHeader = HighDimHeader()
        (length, start) = decoder._DecodeVarint(data, 0)
        hdHeader.ParseFromString(data[start:start+length])
        data = data[start+length:]
        hdRows = []
        n = len(data)
        start = 0
        while start < n:
            (length, start) = decoder._DecodeVarint(data, start)
            hdRow = Row()
            hdRow.ParseFromString(data[start:start+length])
            hdRows.append(hdRow)
            start += length
        return (hdHeader, hdRows)

    def _get_protobuf(self, url):
        headers = {
            'Accept': 'application/octet-stream'
        }
        if self.access_token is not None:
            headers['Authorization'] = 'Bearer ' + self.access_token
        req = urllib.request.Request(url, headers=headers)
        return self._parse_protobuf(urllib.request.urlopen(req).read())





from .api.v1.tm_api_v1 import TransmartV1
from .api.v2.tm_api_v2 import TransmartV2


def get_api(host, api_version=2, user=None, password=None, print_urls=False):
    """
    Create the python transmart client by providing user credentials.

    :param host: a transmart URL (e.g. http://transmart-test.thehyve.net)
    :param user: if not given, it asks for it.
    :param password: if not given, it asks for it.
    :param api_version: either 1 or 2. Default is 2.
    :param print_urls: print the url of handles being used.
    """
    api_versions = (1, 2)

    if api_version not in api_versions:
        raise ValueError("Not a valid TranSMART API version. Choose from: " + str(api_versions))

    api = TransmartV1 if api_version == 1 else TransmartV2

    return api(host, user=user, password=password, print_urls=print_urls)



"""
* Copyright (c) 2015-2017 The Hyve B.V.
* This code is licensed under the GNU General Public License,
* version 3.
"""

import json


class Query:
    """ Utility to build queries for transmart v2 api. """

    def __init__(self, handle=None, method='GET', params=None, hal=False,
                 in_study=None, in_patientset=None, in_concept=None, in_gene_list=None,
                 in_transcript_list=None):
        self.handle = handle
        self.method = method
        self.hal = hal
        self._params = params or {}

        # Subject constraints
        self.in_study = StudyConstraint(in_study)
        self.in_patientset = PatientSetConstraint(in_patientset)
        self.in_concept = ConceptConstraint(in_concept)

        # Biomarker constraints
        self.in_gene_list = GenesConstraint(in_gene_list)
        self.in_transcript_list = TranscriptConstraint(in_transcript_list)

    @property
    def params(self):
        return self._params

    @params.getter
    def params(self):
        self._params.update(self.get_constraints())
        self._params.update(self.get_biomarker_constraints())
        return self._params

    @property
    def headers(self):
        return {'Accept': 'application/{};charset=UTF-8'.format('hal+json' if self.hal else 'json')}

    def get_constraints(self):
        constraints = ''.join([str(c) for c in (self.in_study, self.in_patientset, self.in_concept) if c.value])

        if constraints:
            return {'constraint': constraints}
        else:
            return {}

    def get_biomarker_constraints(self):
        constraints = ''.join([str(c) for c in (self.in_transcript_list, self.in_gene_list) if c.value])

        if constraints:
            return {'biomarker_constraint': constraints}
        else:
            return {}

    def __repr__(self):
        return "<Query(handle={}, method={}, params={})>".format(self.handle,
                                                                 self.method,
                                                                 self.params)


class Constraint:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return json.dumps({"type": self.type_, self.val_name: self.value})


class StudyConstraint(Constraint):
    type_ = 'study_name'
    val_name = 'studyId'


class PatientSetConstraint(Constraint):
    type_ = 'patient_set'
    val_name = 'patientSetId'


class ConceptConstraint(Constraint):
    type_ = 'concept'
    val_name = 'path'


class BiomarkerConstraint(Constraint):

    def __str__(self):
        return json.dumps({"type": "biomarker",
                           "biomarkerType": self.type_,
                           "params": {"names": self.value}})


class TranscriptConstraint(BiomarkerConstraint):
    type_ = 'transcripts'


class GenesConstraint(BiomarkerConstraint):
    type_ = 'genes'



"""
* Copyright (c) 2015-2017 The Hyve B.V.
* This code is licensed under the GNU General Public License,
* version 3.
"""

import requests

from pandas.io.json import json_normalize

from ..tm_api_base import TransmartAPIBase
from .query import Query
from .data_structures import ObservationSet, ObservationSetHD, TreeNodes, PatientSets, Studies, StudyList


class TransmartV2(TransmartAPIBase):
    """ Connect to tranSMART using Python. """

    def __init__(self, host, user=None, password=None, print_urls=False):
        """
        Create the python transmart client by providing user credentials.

        :param host: a transmart URL (e.g. http://transmart-test.thehyve.net)
        :param user: if not given, it asks for it.
        :param password: if not given, it asks for it.
        :param print_urls: print the url of handles being used.
        """
        super().__init__(host, user, password, print_urls)
        self.studies = None

    def query(self, q):
        """ Perform query using API client using a Query object """

        url = "{}{}".format(self.host, q.handle)
        if self.print_urls:
            print(q)

        headers = q.headers
        headers['Authorization'] = 'Bearer ' + self.access_token

        if q.method.upper() == 'GET':
            r = requests.get(url, params=q.params, headers=headers)
        else:
            r = requests.post(url, json=q.params, headers=headers)

        return r.json()

    def get_observations(self, study=None, patient_set=None, as_dataframe=False):
        """
        Get observations, from the main table in the transmart data model.

        :param study: studyID
        :param patient_set: patient set id
        :param as_dataframe: If True, convert json response to dataframe directly
        :return: dataframe or ObservationSet object
        """

        q = Query(handle='/v2/observations',
                  params={'type': 'clinical'},
                  in_study=study,
                  in_patientset=patient_set)

        observations = ObservationSet(self.query(q))

        if as_dataframe:
            return observations.dataframe

        return observations

    def get_patients(self, study=None, patient_set=None, as_dataframe=False):
        """
        Get patients.

        :param study: studyID
        :param patient_set: patient set id
        :param as_dataframe: If True, convert json response to dataframe directly
        :return: dataframe or direct json
        """
        q = Query(handle='/v2/patients', in_study=study, in_patientset=patient_set)

        patients = self.query(q)

        if as_dataframe:
            patients = json_normalize(patients['patients'])
        return patients

    def get_studies(self, as_dataframe=False):
        """
        Get all studies.

        :param as_dataframe: If True, convert json response to dataframe
        :return: dataframe or Studies object
        """

        q = Query(handle='/v2/studies')

        studies = Studies(self.query(q))

        self.studies = StudyList(studies.dataframe.studyId)

        if as_dataframe:
            studies = studies.dataframe

        return studies

    def get_concepts(self, study, hal=False):
        raise NotImplementedError("Call not available for API V2.")

    def tree_nodes(self, root=None, depth=0, counts=True, tags=True, hal=False):
        """

        :param root:
        :param depth:
        :param counts:
        :param tags:
        :param hal:
        :return:
        """
        q = Query(handle='/v2/tree_nodes',
                  params={'root': root,
                          'depth': depth,
                          'counts': counts,
                          'tags': tags},
                  hal=hal)

        return TreeNodes(self.query(q))

    def get_patient_sets(self):
        q = Query(handle='/v2/patient_sets')
        return PatientSets(self.query(q))

    def get_hd_node_data(self, study, hd_type='autodetect', genes=None, transcripts=None, concept=None,
                         patient_set=None, projection='all_data'):
        """

        :param study:
        :param hd_type:
        :param genes:
        :param transcripts:
        :param concept:
        :param patient_set:
        :param projection: ['all_data', 'zscore', 'log_intensity']
        :return:
        """
        q = Query(handle='/v2/observations',
                  method='POST',
                  params={'type': hd_type,
                          'projection': projection},
                  in_study=study,
                  in_patientset=patient_set,
                  in_concept=concept,
                  in_gene_list=genes,
                  in_transcript_list=transcripts
                  )
        return ObservationSetHD(self.query(q))
